###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = 'Donovan Parks'
__copyright__ = 'Copyright 2015'
__credits__ = ['Donovan Parks']
__license__ = 'GPL3'
__maintainer__ = 'Donovan Parks'
__email__ = 'donovan.parks@gmail.com'

from abstract_plot import AbstractPlot
from color_brewer import ColorBrewer


class PieChart(AbstractPlot):
    """Generate pie chart."""

    def __init__(self, options=None):
        """Initialization."""
        AbstractPlot.__init__(self, options)

    def plot(self):  # sizes, labels, min_size=1, other_label='Other'):
        # set size of figure
        self.fig.clear()
        self.fig.set_size_inches(self.options.width, self.options.height)
        axes = self.fig.add_subplot(111, aspect='equal')

        # create plot
        labels = 'Frogs', 'Hogs', 'Dogs', 'Logs'
        sizes = [15, 30, 45, 10]
        colours = ColorBrewer().maps['qualPastel1'].values()[0:len(sizes)]

        axes.pie(sizes,
                 labels=labels,
                 colors=colours,
                 autopct='%1.1f%%',
                 shadow=False,
                 startangle=90)

        self.draw()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################


from biolib.common import is_float


'''Helper functions for parsing Newick information.'''


def parse_label(label):
    """Parse a Newick label which may contain a support value, taxon, and/or auxiliary information.

    Parameters
    ----------
    label : str
        Internal label in a Newick tree.

    Returns
    -------
    float
        Support value specified by label, or None
    str
        Taxon specified by label, or None
    str
        Auxiliary information, on None
    """

    support = None
    taxon = None
    auxiliary_info = None
    
    if label:
        label = label.strip()
        if '|' in label:
            label, auxiliary_info = label.split('|')

        if ':' in label:
            support, taxon = label.split(':')
            support = float(support)
        else:
            if is_float(label):
                support = float(label)
            elif label != '':
                taxon = label

    return support, taxon, auxiliary_info
    
    
def create_label(support, taxon, auxiliary_info):
    """Create label for Newick tree.
    
    Parameters
    ----------
    float
      Support value specified by label, or None
    str
      Taxon specified by label, or None
    str
      Auxiliary information, on None
        
    Returns
    -------
    str
      Valid newick label.
    """
    
    label = ''
    if support is not None and taxon:
        label = str(support) + ':' + taxon
    elif support is not None:
        label = str(support)
    elif taxon:
        label = taxon
        
    if auxiliary_info:
        label += '|' + auxiliary_info

    return label



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = 'Donovan Parks'
__copyright__ = 'Copyright 2014'
__credits__ = ['Donovan Parks', 'Connor Skennerton']
__license__ = 'GPL3'
__maintainer__ = 'Donovan Parks'
__email__ = 'donovan.parks@gmail.com'

import os
import types
import tempfile
import argparse


class ChangeTempAction(argparse.Action):
    """Action for changing the directory used for temporary files.

    Example:
        <parse>.add_argument('--tmpdir', action=ChangeTempAction, default=tempfile.gettempdir(), help="specify alternative directory for temporary files")
    """
    def __call__(self, parser, namespace, values, option_string=None):
        if os.path.isdir(values):
            tempfile.tempdir = values
            setattr(namespace, self.dest, values)
        else:
            raise argparse.ArgumentTypeError('The value of %s must be a valid directory' % option_string)


class CustomHelpFormatter(argparse.HelpFormatter):
    """Provide a customized format for help output.

    http://stackoverflow.com/questions/9642692/argparse-help-without-duplicate-allcaps
    """

    def _get_help_string(self, action):
        """Place default value in help string."""
        h = action.help
        if '%(default)' not in action.help:
            if action.default != '' and action.default != [] and action.default is not None and type(action.default) != types.BooleanType:
                if action.default is not argparse.SUPPRESS:
                    defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]

                    if action.option_strings or action.nargs in defaulting_nargs:
                        if '\n' in h:
                            lines = h.splitlines()
                            lines[0] += ' (default: %(default)s)'
                            h = '\n'.join(lines)
                        else:
                            h += ' (default: %(default)s)'
            return h

    def _format_action_invocation(self, action):
        """Removes duplicate ALLCAPS with positional arguments."""
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            metavar, = self._metavar_formatter(action, default)(1)
            return metavar

        else:
            parts = []

            # if the Optional doesn't take a value, format is:
            #    -s, --long
            if action.nargs == 0:
                parts.extend(action.option_strings)

            # if the Optional takes a value, format is:
            #    -s ARGS, --long ARGS
            else:
                default = self._get_default_metavar_for_optional(action)
                args_string = self._format_args(action, default)
                for option_string in action.option_strings:
                    parts.append(option_string)

                return '%s %s' % (', '.join(parts), args_string)

            return ', '.join(parts)

    def _get_default_metavar_for_optional(self, action):
        return action.dest.upper()

    def _get_default_metavar_for_positional(self, action):
        return action.dest



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = 'Donovan Parks'
__copyright__ = 'Copyright 2015'
__credits__ = ['Donovan Parks']
__license__ = 'GPL3'
__maintainer__ = 'Donovan Parks'
__email__ = 'donovan.parks@gmail.com'

from abstract_plot import AbstractPlot

from matplotlib.ticker import NullFormatter, FuncFormatter

from numpy import (zeros_like as np_zeros_like,
                    ones_like as np_ones_like,
                    array as np_array)

def to_percent(y, position):
    # Ignore the passed in position. This has the effect of scaling the default
    # tick locations.
    s = str(100 * y)

    return s

class Histogram(AbstractPlot):
    """Histogram plot."""

    def __init__(self, options):
        AbstractPlot.__init__(self, options)

    def plot(self, x, 
                    xlabel, 
                    ylabel, 
                    bins, 
                    color, 
                    yticks=None, 
                    xticks=None, 
                    xlim=None, 
                    invert_xaxis=False,
                    stacked=False,
                    show_legend=False):
        # set size of figure
        self.fig.clear()
        self.fig.set_size_inches(self.options.width, self.options.height)
        axis = self.fig.add_subplot(111)

        # plot histogram as percentages
        num_entries = sum([len(c) for c in x])
        weights = []
        for c in xrange(0, len(x)):
            w = [100.0/num_entries for _ in xrange(0, len(x[c]))]
            weights.append(w)
        
        #weights = []
        #for c in xrange(0, len(x)):
        #    num_entries = len(x[c])
        #    w = [100.0/num_entries for _ in xrange(0, len(x[c]))]
        #    weights.append(w)

        pdf, bins, patches = axis.hist(x, bins=bins, 
                                            #normed=normed,
                                            weights=weights,
                                            color=color, 
                                            lw=0.5, 
                                            histtype='bar', 
                                            stacked=stacked)
        
        axis.set_xlabel(xlabel)
        axis.set_ylabel(ylabel)
        
        #formatter = FuncFormatter(to_percent)
        #axis.yaxis.set_major_formatter(formatter)
        
        if show_legend:
            lgnd = axis.legend(
                               loc='upper left',
                               ncol=1,
                               fontsize=self.options.tick_font_size,
                               handletextpad=0.5,
                               markerscale=2,
                               frameon=False)
        
        # *** Prettify histogram plot
        if yticks:
            axis.set_yticks(yticks)

        if xlim:
            axis.set_xlim(xlim)
        
        if xticks:
            axis.set_xticks(xticks)
            
        if invert_xaxis:
            axis.invert_xaxis()

        self.prettify(axis)

        self.fig.tight_layout(pad=1.0, w_pad=0.1, h_pad=0.1)
        self.draw()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Donovan Parks"
__copyright__ = "Copyright 2017"
__credits__ = ["Donovan Parks"]
__license__ = "GPL3"
__maintainer__ = "Donovan Parks"
__email__ = "donovan.parks@gmail.com"
__status__ = "Development"

import os
import logging
import time
import shutil
import tempfile
import ntpath

import biolib.seq_io as seq_io
from biolib.parallel import Parallel
from biolib.bootstrap import bootstrap_alignment, bootstrap_support
from biolib.external.execute import check_on_path


class RAxML():
    """Wrapper for running RAxML."""

    def __init__(self, cpus):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

        self.cpus = cpus

        if self.cpus > 1:
            check_on_path('raxmlHPC-PTHREADS-SSE3')
        else:
            check_on_path('raxmlHPC-SSE3')

    def bootstrap(self, input_tree, msa_file, model_str, num_replicates, output_dir, cpus):
        """Perform non-parametric bootstrapping.

        Parameters
        ----------
        input_tree : str
            File containing newick tree to decorate with bootstraps.
        msa_file : str
            Fasta file containing multiple sequence alignment.
        model_str : str
            Specified either the 'WAG' or 'LG' model.
        num_replicates : int
            Number of replicates to perform.
        output_dir: str
            Output directory to contain bootstrap trees.
        cpus : int
            Number of cpus to use.
        """
        
        check_on_path('seqmagick')

        assert(model_str.upper() in ['WAG', 'LG'])

        self.output_dir = output_dir
        self.model = model_str
        self.msa = seq_io.read(msa_file)

        # calculate replicates
        parallel = Parallel(cpus)
        parallel.run(self._bootstrap, None, xrange(num_replicates), None)

        # calculate support values
        rep_tree_files = []
        for rep_index in xrange(num_replicates):
            rep_tree_files.append(os.path.join(output_dir, 'rep_%d' % rep_index, 'RAxML_bestTree.support'))

        tree_name = os.path.splitext(os.path.basename(input_tree))[0]
        output_tree = os.path.join(output_dir, tree_name + '.bootstrap.tree')
        bootstrap_support(input_tree, rep_tree_files, output_tree)
        
        return output_tree

    def _bootstrap(self, rep_num):
        """Infer tree from bootstrapped multiple sequence alignment.

        Parameters
        ----------
        replicated_num : int
          Unique replicate number.
        cpus : int
            Number of cpus to use.
        """
        
        rep_dir = os.path.join(self.output_dir, 'rep_%d' % rep_num)
        if not os.path.exists(rep_dir):
            os.makedirs(rep_dir)

        output_msa = os.path.join(rep_dir, 'bootstrap.afa')
        bootstrap_alignment(self.msa, output_msa)
        
        phylip_msa_file = output_msa.replace('.afa', '.phyx')
        cmd = 'seqmagick convert %s %s' % (output_msa, phylip_msa_file)
        os.system(cmd)

        log_file = os.path.join(rep_dir, 'raxml.log')
        self.run(phylip_msa_file, self.model, rep_dir, log_file, output_suffix='support')

        return True

    def run(self, msa_file, model_str, output_dir, log_file=None, output_suffix=None):
        """Infer tree using RAxML.

        All trees are inferred using the GAMMA distribution to model
        rate heterogeneity. 

        Parameters
        ----------
        msa_file : str
            Phylip file containing multiple sequence alignment.
        model_str : str
            Specified either the 'WAG', 'LG', or 'AUTO' model.
        output_dir: str
            Output directory for inferred tree and auxillary information.
        log_file: str
            Output file containing information about inferred tree.
            
        Returns
        -------
        str
            Path to inferred tree.
        """

        assert(model_str.upper() in ['WAG', 'LG', 'AUTO'])

 
        output_dir = os.path.abspath(output_dir)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        
        if not log_file:
            log_file = os.path.join(output_dir, 'raxml.log')
            
        if not output_suffix:
            output_suffix = time.strftime("%Y%m%d")
            
        model_str = 'PROTGAMMA%s' % model_str.upper()
        cmd = '-s %s -m %s -p 12345 -w %s -n %s > %s' % (msa_file,
                                                            model_str,
                                                            output_dir,
                                                            output_suffix,
                                                            log_file)
        if self.cpus > 1:
            cmd = 'raxmlHPC-PTHREADS-SSE3 -T %d ' % self.cpus + cmd
        else:
            cmd = 'raxmlHPC-SSE3 ' + cmd
        os.system(cmd)

        return os.path.join(output_dir, 'RAxML_bestTree.%s' % output_suffix)


import os


def version():
    """Read program version from file."""
    binDir = os.path.dirname(os.path.realpath(__file__))
    versionFile = open(os.path.join(binDir, 'VERSION'))
    return versionFile.read().strip()

__version__ = version()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Donovan Parks"
__copyright__ = "Copyright 2015"
__credits__ = ["Donovan Parks"]
__license__ = "GPL3"
__maintainer__ = "Donovan Parks"
__email__ = "donovan.parks@gmail.com"
__status__ = "Development"

import os
import logging
import shutil
import tempfile
import ntpath

import biolib.seq_io as seq_io
from biolib.parallel import Parallel
from biolib.bootstrap import bootstrap_alignment, bootstrap_support
from biolib.external.execute import check_on_path


class FastTree():
    """Wrapper for running FastTree."""

    def __init__(self, multithreaded=True):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

        self.multithreaded = multithreaded

        if self.multithreaded:
            check_on_path('FastTreeMP')
        else:
            check_on_path('FastTree')

    def bootstrap(self, input_tree, msa_file, seq_type, model_str, num_replicates, output_dir, cpus):
        """Perform non-parametric bootstrapping.

        Parameters
        ----------
        input_tree : str
            File containing newick tree to decorate with bootstraps.
        msa_file : str
            Fasta file containing multiple sequence alignment.
        seq_type : str
            Specifies multiple sequences alignment is of 'nt' or 'prot'.
        model_str : str
            Specified either the 'wag' or 'jtt' model.
        num_replicates : int
            Number of replicates to perform.
        output_dir: str
            Output directory to contain bootstrap trees.
        cpus : int
            Number of cpus to use.
        """

        assert(seq_type.upper() in ['NT', 'PROT'])
        assert(model_str.upper() in ['WAG', 'LG', 'JTT', 'GTR'])

        self.output_dir = output_dir
        self.seq_type = seq_type
        self.model = model_str
        self.msa = seq_io.read(msa_file)

        # calculate replicates
        parallel = Parallel(cpus)
        parallel.run(self._bootstrap, None, xrange(num_replicates), None)

        # calculate support values
        rep_tree_files = []
        for rep_index in xrange(num_replicates):
            rep_tree_files.append(os.path.join(self.output_dir, 'rep_%d' % rep_index, 'bootstrap.tree'))

        tree_name = os.path.splitext(os.path.basename(input_tree))[0]
        output_tree = os.path.join(output_dir, tree_name + '.bootstrap.tree')
        bootstrap_support(input_tree, rep_tree_files, output_tree)
        
        return output_tree

    def _bootstrap(self, rep_num):
        """Infer tree from bootstrapped multiple sequence alignment.

        Parameters
        ----------
        replicated_num : int
          Unique replicate number.
        cpus : int
            Number of cpus to use.
        """
        
        rep_dir = os.path.join(self.output_dir, 'rep_%d' % rep_num)
        if not os.path.exists(rep_dir):
            os.makedirs(rep_dir)

        output_msa = os.path.join(rep_dir, 'bootstrap.afa')
        bootstrap_alignment(self.msa, output_msa)

        output_tree = os.path.join(rep_dir, 'bootstrap.tree')
        log_file = os.path.join(rep_dir, 'fasttree.log')
        self.run(output_msa, self.seq_type, self.model, output_tree, log_file)

        return True

    def parallel_run(self, msa_files, seq_type, model_str, output_dir, cpus):
        """Infer tree using FastTree in parallel.

        Parameters
        ----------
        msa_files : str
            Fasta files containing multiple sequence alignments.
        seq_type : str
            Specifies multiple sequences alignment is of 'nt' or 'prot'.
        model_str : str
            Specified either the 'wag' or 'jtt' model.
        output_dir: str
            Prefix for all output files.
        """

        assert(seq_type.upper() in ['NT', 'PROT'])
        assert(model_str.upper() in ['WAG', 'LG', 'JTT', 'GTR'])

        self.output_dir = output_dir
        self.seq_type = seq_type
        self.model = model_str

        parallel = Parallel(cpus)
        parallel.run(self._parallel_infer_tree, None, msa_files, None)

    def _parallel_infer_tree(self, msa_file):
        """Infer tree using FastTree in parallel.

        Parameters
        ----------
        msa_file : str
            Fasta files containing multiple sequence alignments.
        """

        file_prefix = ntpath.basename(msa_file)
        if '.' in file_prefix:
            file_prefix = file_prefix[0:file_prefix.find('.')]

        output_tree = os.path.join(self.output_dir, file_prefix + '.tree')
        fast_tree_output = os.path.join(self.output_dir, file_prefix + '.log')
        self.run(msa_file, self.seq_type, self.model, output_tree, fast_tree_output)

    def run(self, msa_file, seq_type, model_str, output_tree, output_tree_log, log_file=None):
        """Infer tree using FastTree.

        All trees are inferred using the GAMMA distribution to model
        rate heterogeneity. Nucleotide trees are inferred under the
        GTR model.

        Parameters
        ----------
        msa_file : str
            Fasta file containing multiple sequence alignment.
        seq_type : str
            Specifies multiple sequences alignment is of 'nt' or 'prot'.
        model_str : str
            Specified either the 'wag', 'lg', or 'jtt' model.
        output_tree: str
            Output file containing inferred tree.
        output_tree_log: str
            Output file containing information about inferred tree.
        output_log: str
            Output file containing information about running of FastTree.
        """

        assert(seq_type.upper() in ['NT', 'PROT'])
        assert(model_str.upper() in ['WAG', 'LG', 'JTT', 'GTR'])

        if seq_type == 'prot':
            seq_type_str = ''
            if model_str.upper() == 'JTT':
                model_str = ''
            elif model_str.upper() == 'WAG':
                model_str = '-wag'
            elif model_str.upper() == 'LG':
                model_str = '-lg'
        elif seq_type == 'nt':
            seq_type_str = '-nt'
            model_str = '-gtr'

        if not log_file:
            log_file = '/dev/null'

        cmd = '-quiet -nosupport -gamma %s %s -log %s %s > %s 2> %s' % (seq_type_str,
                                                                        model_str,
                                                                        output_tree_log,
                                                                        msa_file,
                                                                        output_tree,
                                                                        log_file)
        if self.multithreaded:
            cmd = 'FastTreeMP ' + cmd
            os.system(cmd)
        else:
            cmd = 'FastTree ' + cmd
            os.system(cmd)
